# -*- coding: utf-8 -*-
"""
    Module: aypylib:math
    File:   fit_Gauss.py
    Descr.: SciPy.optimize.curve_fit utility
            Gaussian fitting
    Update: 200507 v.1.00 build as xwFitGa.py
            210228 v.1.01 rebuild as fitting lib.
            211011 v.1.10 add param. fnFit(), kwd-args化
    Last:   220213 v.1.11 fitGa() RuntimeError handle: sopt.curve_fit()
@author: Ayoshida(RIKEN,RNC)
"""

""" import: standard lib.       
----------------------------------------------      """
import numpy             as np
import scipy.optimize    as sopt

"""
    Global var. in this module
"""
_DBG  = False   # for debug msg. CTRL

""" ==============================================================
        Fitting func.
    ============================================================== """
# gaussian function
def fn_gaussian(x:list, xm:float, sgm:float, a:float):
    return( a * np.exp( - (x - xm)**2 / (2 * sgm**2)) )

""" ==============================================================
        Module func.
    ============================================================== """
def fitGa_ini(x, y):
    """ Find rough initial value of [xm,sgm,a]
    x,y:    1d-ndarray    X-,Y-data
    <Ret.>
    pini:   list          [xm,sgm,a] result
    ----------------------------------------------      """
    # [仮定]　x,yの範囲内には、peakが１つのみ　あるとする　
    # np.mean() を用いて ラフ な pini を決める
    # yの最大値:a の a1=a*exp(-1/2):x1=xm±sgm 以上の部分の mean-value を xm とする 
    a  = max(y)
    a1 = a*np.exp(-1/2)
    x1 = x[np.where(y >= a1)]
    xm = np.mean(x1)
    # sgm= (max(x1) - min(x1))/2 とする。
    sgm = (max(x1) - min(x1))/2
    if(_DBG): print('_fit_ini: a=',a,' len(x1)=',len(x1),' xm=',xm,' sgm=',sgm)
    pini= [xm,sgm,a]
    return( pini )

def fitGa(x, y,*,
        xm0:float=None, sg0:float=None, a0:float=None,
        fnFit:callable=None ) -> list:
    """ Gaussian fitting using SciPy.optimize.curve_fit :
        func.: Y(X) = a * exp( - (X - xm)**2 / (2 * sgm**2)) 
    x,y:    1d-ndarray    X-,Y-data
    <kwd args>
    xm0:    float       init xm  value
    sg0:    float       init sgm value
    a0:     float       init a   value
    fnFit:  callable    fit model func fnFit(x,xm,sg,a) 
        defa) fn_gaussian(x,xm,sg,a) 
        cf-J) fnFit() のパラメータは x,xm,sg,a に限ります。
              Gaussian + const.B.G. とか、
              Gaussian + B.G(Ax+B) A,B:const 等
    <Ret.>  fitting result
    popt:   list        [xm,sgm,a,rsq]
    yfit:   1d-ndarray  resultant gaussian along with x; fit結果プロット用
    ----------------------------------------------      """
    if((xm0 is None) or (sg0 is None) or (a0 is None)):
        pini = fitGa_ini(x,y)
    else:
        pini = [xm0,sg0,a0]
    if(_DBG): print('pini=',pini)
    _fn = fn_gaussian if(fnFit is None) else fnFit
    try:
        popt, pcov = sopt.curve_fit(_fn, x, y, pini)
    except RuntimeError as e:
        print(e)
        return([0,0,0,0],None)
    [xm,sg,a] = popt
    yfit= _fn(x,*popt)      # resultant gaussian along with x
    res= y - yfit                       # residuals
    rss= np.sum( res**2 )               # residual sum of squares
    tss= np.sum((y - np.mean(y))**2)    # total sum of squares
    rsq= 1 - (rss / tss)                # R^2 決定係数
    if(_DBG): print('popt=',popt,'R2=',rsq)
    return( [xm,sg,a,rsq], yfit )

def fitGaRoi(x, y,*,
        xroi1:float=None, xroi2:float=None,
        xm0:float=None, sg0:float=None, a0:float=None,
        fnFit:callable=None ) -> list:
    """ Gaussian fitting using SciPy.optimize.curve_fit :
        func.: Y(X) = a * exp( - (X - xm)**2 / (2 * sgm**2))
        with ROI for X-,Y-data
    x,y:    1d-ndarray    X-,Y-data
    <kwd args>
    xroi1:  float       X-data ROI low
    xroi2:  float       X-data ROI high
    xm0:    float       init xm  value
    sg0:    float       init sgm value
    a0:     float       init a   value
    fnFit:  callable    fit model func fnFit(x,xm,sg,a) 
   <Ret.>  fitting result
    popt:   list        [xm,sgm,a,rsq] 
    xfit:   1d-ndarray  resultant X-data ROI 
    yfit:   1d-ndarray  resultant gaussian along with xfit; fit結果プロット用
    ----------------------------------------------      """
    xx, yy = np.array(x), np.array(y)
    if((xroi1 is not None) and (xroi2 is not None)):
        ix = np.where((xx >= xroi1) & (xx <= xroi2))
        xx,yy = xx[ix],yy[ix]
        if(_DBG): print('xroi=',xroi1,',',xroi2)
    popt,yfit = fitGa(xx,yy,xm0=xm0,sg0=sg0,a0=a0,fnFit=fnFit)
    return( popt,xx,yfit )

""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    import matplotlib.pyplot as plt

    SwTest01 = False     # test xwfitGa() auto
    SwTest02 = False     # test xwfitGa() manual 
    SwTest03 = False     # test xwfitGaRoi() auto
    SwTest04 = False     # test xwfitGaRoi() manual
    SwTest13 = True      # test xwfitGaRoi() manual
    
    """ Prepare test data
        ----------------------------------------------      """
    x0,x1,nx   =  0,100,200     # X binning
    xm1,sg1,a1 = 30, 10,100     # Gaussian1
    xm2,sg2,a2 = 60,  5,150     # Gaussian2
    x_bin  = np.linspace(x0, x1, nx)
    y_noise= 0.5*np.random.normal(size=nx)
    y_ga1  = fn_gaussian(x_bin, xm1, sg1, a1) + y_noise
    y_noise= 5.0*np.random.normal(size=nx)
    y_ga2  = fn_gaussian(x_bin, xm2, sg2, a2) + y_noise
    y_dat  = y_ga1 + y_ga2
    if(SwTest13):
        #Slp,BG = 1.5, 10
        Slp,BG = 1.5, 50
        y_bg = np.full_like(x_bin,BG)
        y_bg = Slp * x_bin + BG
        y_dat += y_bg

    plt.plot(x_bin,y_ga1,color='green', linestyle='dashed')
    plt.plot(x_bin,y_ga2,color='blue',  linestyle='dashed')
    plt.plot(x_bin,y_dat,color='orange',linestyle='dashed')

    """ Test01: xwfitGa() auto mode
        ----------------------------------------------      """
    if(SwTest01):
        [xm,sg,a,R2],yfit =  fitGa(x_bin, y_dat)

    """ Test02: xwfitGa() manual mode
        ----------------------------------------------      """
    if(SwTest02):
        pini = {'xm0':xm2+3, 'sg0':sg2+5, 'a0':a2-20}
        # values = list(pini.values())  dict。values() -> list 変換
        [xm,sg,a,R2],yfit =  fitGa(x_bin, y_dat, **pini)

    """ Test03: xwfitGaRoi() auto mode
        ----------------------------------------------      """
    if(SwTest03):
        xroi1,xroi2 = xm2-2*sg2, xm2+2*sg2
        #xroi1,xroi2 = xm2-1.5*sg2, xm2+0.9*sg2  # +/- > 1.1*sgm 必要らしい
        #xroi1,xroi2 = x1,x0        # Error check1
        [xm,sg,a,R2],x_bin,yfit =  fitGaRoi(x_bin, y_dat, 
                                xroi1=xroi1, xroi2=xroi2)

    """ Test04: xwfitGaRoi() manual mode
        ----------------------------------------------      """
    if(SwTest04):
        xroi1,xroi2 = xm2-2*sg2, xm2+2*sg2
        #xroi1,xroi2 = xm2-1.5*sg2, xm2+0.9*sg2
        pini = {'xm0':xm2+3, 'sg0':sg2+5, 'a0':a2-20}
        [xm,sg,a,R2],x_bin,yfit =  fitGaRoi(x_bin, y_dat,
                    xroi1=xroi1, xroi2=xroi2, **pini)

    """ Test13: fnFit()0 + xwfitGaRoi() auto mode
        ----------------------------------------------      """
    if(SwTest13):
        
        def fn_fit(x:list, xm:float, sgm:float, a:float):
            #print('Slp,BG=',Slp,BG)
            return( a * np.exp( - (x - xm)**2 / (2 * sgm**2)) 
                   + (Slp * x + BG) )

        plt.plot(x_bin,y_bg,color='cyan',  linestyle='solid')
        xroi1,xroi2 = xm2-3*sg2, xm2+3*sg2
        #xroi1,xroi2 = xm2-1.5*sg2, xm2+0.9*sg2  # +/- > 1.1*sgm 必要らしい
        #xroi1,xroi2 = x1,x0        # Error check1
        if(True):
            [xm,sg,a,R2],x_bin,yfit =  fitGaRoi(x_bin, y_dat, 
                    xroi1=xroi1, xroi2=xroi2, fnFit=fn_fit)
        else:
            [xm,sg,a,R2],x_bin,yfit =  fitGaRoi(x_bin, y_dat, 
                    xroi1=xroi1, xroi2=xroi2)

    plt.plot(x_bin,yfit,color='red',linestyle='solid')
    ms = 'xm={:.2f}, sg={:.2f}, a={:.2f}, R2={:.4f}'.format(
        xm,sg,a,R2)
    print(ms)
    print('@@main: -- Fin ---')

    
""" References: ============================================

* scipy.optimize.curve_fit 
  https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html#scipy.optimize.curve_fit
* SciPyフィッティング (fitting)　Yamamoto's Laboratory 
  http://www.yamamo10.jp/yamamoto/comp/Python/library/SciPy/fit/index.php
* Qiita2018 研究者のための実践データ処理～Pythonでピークフィッティング～
  https://qiita.com/kon2/items/6498e66af55949b41a99
* サボテンパイソン Scipyのcurve_fitで最小2乗法近似、決定係数R2も求める
  https://sabopy.com/py/curve_fit/
* 高校数学の美しい物語 共分散の意味と簡単な求め方
  https://mathtrain.jp/covariance
* 高校数学の美しい物語 決定係数の定義と相関係数との関係
  https://mathtrain.jp/ketteikeisu
 ============================================================ """
