# -*- coding: utf-8 -*-
"""
    Module: aypylib:xlwings
    File:   xwFitGa.py
    Descr.: SciPy.optimize.curve_fit utility
            Gaussian fitting
    Update: 200507 v.1.00 build
@author: Ayoshida(RIKEN,RNC)
"""

""" import: standard lib.       
----------------------------------------------      """
import xlwings           as xw
import numpy             as np
import scipy.optimize    as sopt

"""
    Global var. in this module
"""
_DBG  = False    # for debug msg. CTRL

""" ==============================================================
        Fitting func.
    ============================================================== """
# gaussian function
def fit_gaussian(x, xm, sgm, a):
    return( a * np.exp( - (x - xm)**2 / (2 * sgm**2)) )

""" ==============================================================
        Helper func.
    ============================================================== """
def _fit_ini(x, y):
    """ Find rough initial value of [xm,sgm,a]
    x,y:    1d-ndarray    X-,Y-data
    <Ret.>
    pini:   [xm,sgm,a]    result
    ----------------------------------------------      """
    # [仮定]　x,yの範囲内には、peakが１つのみ　あるとする　
    # np.mean() を用いて ラフ な pini を決める
    # yの最大値:a の a1=a*exp(-1/2):x1=xm±sgm 以上の部分の mean-value を xm とする 
    a  = max(y)
    a1 = a*np.exp(-1/2)
    x1 = x[np.where(y >= a1)]
    xm = np.mean(x1)
    # sgm= (max(x1) - min(x1))/2 とする。
    sgm = (max(x1) - min(x1))/2
    if(_DBG): print('_fit_ini: a=',a,' len(x1)=',len(x1),' xm=',xm,' sgm=',sgm)
    pini= [xm,sgm,a]
    return( pini )

def _fit_main(x, y, xm0, sg0, a0):
    """ Do curve fit and return [xm,sg,a,R2]
    x,y:    1d-ndarray    X-,Y-data
    xm0:    float       init xm  value
    sg0:    float       init sgm value
    a0:     float       init a   value
    <Ret.>
    [xm,sgm,a,R2]: tuple of float
    ----------------------------------------------      """
    if((xm0 is None) or (sg0 is None) or (a0 is None)):
        pini = _fit_ini(x,y)
    else:
        pini = [xm0,sg0,a0]
    if(_DBG): print('pini=',pini)
    popt, pcov = sopt.curve_fit(fit_gaussian, x, y, pini)
    [xm,sg,a] = popt
    res= y - fit_gaussian(x,xm,sg,a)    # residuals
    rss= np.sum( res**2 )               # residual sum of squares
    tss= np.sum((y - np.mean(y))**2)    # total sum of squares
    rsq= 1 - (rss / tss)                # R^2 決定係数
    if(_DBG): print('popt=',popt,'R2=',rsq)
    return( [xm,sg,a,rsq] )

""" ========================================================
            User Defined Func (UDF)
    ========================================================"""
@xw.func
@xw.arg('x', doc='X-data array to fit')
@xw.arg('y', doc='Y-data array to fit')
@xw.arg('xm0', doc='init xm  value : optional')
@xw.arg('sg0', doc='init sgm value : optional')
@xw.arg('a0',  doc='init a   value : optional')
@xw.ret(transpose=True)
def xwfitGa(x, y,
        xm0:float=None, sg0:float=None, a0:float=None) -> list:
    """ Gaussian fitting using SciPy.optimize.curve_fit :
        func.: Y(X) = a * exp( - (X - xm)**2 / (2 * sgm**2)) 
    x:      range       X-data array to fit
    y:      range       Y-data array to fit
    <optional>
    xm0:    float       init xm  value
    sg0:    float       init sgm value
    a0:     float       init a   value
    <Ret.>  fitting result
    popt:   list:transpose  = [xm,sgm,a] 
    ----------------------------------------------      """
    xx, yy = np.array(x), np.array(y)
    return( _fit_main(xx,yy,xm0,sg0,a0) )

@xw.func
@xw.arg('x', doc='X-data array to fit')
@xw.arg('y', doc='Y-data array to fit')
@xw.arg('xroi1', doc='X-data ROI low:  optional')
@xw.arg('xroi2', doc='X-data ROI high: optional')
@xw.arg('xm0', doc='init xm  value : optional')
@xw.arg('sg0', doc='init sgm value : optional')
@xw.arg('a0',  doc='init a   value : optional')
@xw.ret(transpose=True)
def xwfitGaRoi(x, y,
        xroi1:float=None, xroi2:float=None,
        xm0:float=None, sg0:float=None, a0:float=None) -> list:
    """ Gaussian fitting using SciPy.optimize.curve_fit :
        func.: Y(X) = a * exp( - (X - xm)**2 / (2 * sgm**2))
        with ROI for X-,Y-data
    x:      range       X-data array to fit
    y:      range       Y-data array to fit
    <optional>
    xroi1:  float       X-data ROI low
    xroi2:  float       X-data ROI high
    xm0:    float       init xm  value
    sg0:    float       init sgm value
    a0:     float       init a   value
    <Ret.>  fitting result
    popt:   list:transpose  = [xm,sgm,a] 
    ----------------------------------------------      """
    xx, yy = np.array(x), np.array(y)
    if((xroi1 is not None) and (xroi2 is not None)):
        ix = np.where((xx >= xroi1) & (xx <= xroi2))
        xx,yy = xx[ix],yy[ix]
        if(_DBG): print('xroi=',xroi1,',',xroi2)
    return( _fit_main(xx,yy,xm0,sg0,a0) )

""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':

    import matplotlib.pyplot as plt

    SwTest1 = False     # test xwfitGa() auto
    SwTest2 = False     # test xwfitGa() manual 
    SwTest3 = True      # test xwfitGaRoi() auto
    SwTest4 = False     # test xwfitGaRoi() manual
    
    """ Prepare test data
        ----------------------------------------------      """
    x0,x1,nx   =  0,100,200     # X binning
    xm1,sg1,a1 = 30, 10,100     # Gaussian1
    xm2,sg2,a2 = 60,  5,150     # Gaussian2
    x_bin  = np.linspace(x0, x1, nx)
    y_noise= 0.5*np.random.normal(size=nx)
    y_ga1  = fit_gaussian(x_bin, xm1, sg1, a1) + y_noise
    y_noise= 5.0*np.random.normal(size=nx)
    y_ga2  = fit_gaussian(x_bin, xm2, sg2, a2) + y_noise
    y_dat  = y_ga1 + y_ga2
    #y_dat  = y_ga2
    
    plt.plot(x_bin,y_ga1,color='green', linestyle='dashed')
    plt.plot(x_bin,y_ga2,color='blue',  linestyle='dashed')
    plt.plot(x_bin,y_dat,color='orange',linestyle='dashed')

    """ Test1: xwfitGa() auto mode
        ----------------------------------------------      """
    if(SwTest1):
        [xm,sg,a,R2] =  xwfitGa(x_bin, y_dat)

    """ Test2: xwfitGa() manual mode
        ----------------------------------------------      """
    if(SwTest2):
        pini = [xm2+3, sg2+5, a2-20]
        [xm,sg,a,R2] =  xwfitGa(x_bin, y_dat, pini)

    """ Test3: xwfitGaRoi() auto mode
        ----------------------------------------------      """
    if(SwTest3):
        xroi1,xroi2 = xm2-2*sg2, xm2+2*sg2
        #xroi1,xroi2 = xm2-1.5*sg2, xm2+0.9*sg2  # +/- > 1.1*sgm 必要らしい
        #xroi1,xroi2 = x1,x0        # Error check1
        [xm,sg,a,R2] =  xwfitGaRoi(x_bin, y_dat, xroi1, xroi2)

    """ Test4: xwfitGaRoi() manual mode
        ----------------------------------------------      """
    if(SwTest4):
        xroi1,xroi2 = xm2-2*sg2, xm2+2*sg2
        #xroi1,xroi2 = xm2-1.5*sg2, xm2+0.9*sg2
        pini = [xm2+3, sg2+5, a2-20]
        [xm,sg,a,R2] =  xwfitGaRoi(x_bin, y_dat, xroi1, xroi2, pini)

    y_fit = fit_gaussian(x_bin, xm, sg, a)
    plt.plot(x_bin,y_fit,color='red',linestyle='solid')

    print('@@main: -- Fin ---')

    
""" References: ============================================
* xlwings: VBA: ユーザー定義関数 (UDFs)
  https://docs.xlwings.org/ja/latest/udfs.html
* Qiita 2019 ExcelにPythonが搭載？その後 - xlwings を使おう
　 https://qiita.com/yniji/items/b38bc312e860027108ac
* HatenaBlog 2019 VBAからPythonのコードを実行してみる(1,2)
　 https://celaeno42.hatenablog.com/entry/2018/12/03/191854
　 https://celaeno42.hatenablog.com/entry/2018/12/07/180351
* Newton Excel Bach, not (just) an Excel Blog
  xlwSciPy update for Python 3; April 9, 2018 
  https://newtonexcelbach.com/2018/04/09/xlwscipy-update-for-python-3/

* scipy.optimize.curve_fit 
  https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html#scipy.optimize.curve_fit
* SciPyフィッティング (fitting)　Yamamoto's Laboratory 
  http://www.yamamo10.jp/yamamoto/comp/Python/library/SciPy/fit/index.php
* Qiita2018 研究者のための実践データ処理～Pythonでピークフィッティング～
  https://qiita.com/kon2/items/6498e66af55949b41a99
* サボテンパイソン Scipyのcurve_fitで最小2乗法近似、決定係数R2も求める
  https://sabopy.com/py/curve_fit/
* 高校数学の美しい物語 共分散の意味と簡単な求め方
  https://mathtrain.jp/covariance
* 高校数学の美しい物語 決定係数の定義と相関係数との関係
  https://mathtrain.jp/ketteikeisu
 ============================================================ """
