# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   srimWS.py
    Descr.: def. class srimWS ; Work Sheet
    Update: 220403 build start Ay
            220515 v.1.0 done
    Last:   230424 v.1.1 refactored & add Error handle

    【注】 class srimWS を変更したら、.pickle も作り直す事！

@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  numpy   as  np

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    pySRIMfit   import  (_Err, atomSym, srimWS_Uid, 
                srimWSname_split, srimWSdump, 
                srimWSplot_LET, srimWSplot_RngSt)

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class srimWS
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
    
class srimWS( object ):
    """ SRIM Work-Sheet (srWS) container
Params. for constructor:
    non
Attributes:
  < sheet-info(1): basic >
    .wsName     str   srimWS name
    .SRIMver    str   SRIM version
    .IonZ       int   Beam Z-number
    .IonA       int   Beam A-number
    .Beam       str   eg) '84Kr'
    .Target     str   Target material (short name)
    .TargetNm   str   Target material (long name)
    .TrgDens    float [g/cm3]       Target density
    .TrgDensAtm float [atoms/cm3]   Target density
    .BraggC     float [%]           Bragg corection
    .Gas        bool        Target is Gas ?
    .GasPtbl    float [Pa]   if Gas, its Pressure asuumed in this SRIM table.
    .GasTtbl    float [degC] if Gas, its Temperature asuumed.
    .Corded     str   Corded by info.
  < internal tables >
    ._E         ndary [MeV/A]   E table
    ._LETe      ndary [MeV/(mg/cm2)] dE/dX_elec. table
    ._LETn      ndary [MeV/(mg/cm2)] dE/dX_nucl. table
    ._LETt      ndary [MeV/(mg/cm2)] dE/dX_tot = elec+nucl table
    ._LETe{L|H} (ndary,ndary) (E[] vs LETe[]) at Lower/Higher-side of its peak
    ._LETn{L|H} (ndary,ndary) (E[] vs LETn[]) ''
    ._LETt{L|H} (ndary,ndary) (E[] vs LETt[]) ''
    ._Rng       ndary [um] Range table
    ._StLng     ndary [um] Straggling Longitudinal table
    ._StLtr     ndary [um] Straggling Lateral table
    ._CnvF      ndary      LET unit conversion factor table
    ._TrgCmp    [{}]       Target material composition table like;
                [{'AtomSym':'C','AtomNum':6,'Atom%':1.,'Mass%':1.},..]
  < unit used below >
     Energy     [MeV/A], 
     Range      [um]
     Straggling [um]
     LET{e|n|t} [MeV/(mg/cm2)](Uid=0 defa.); deppends on Uid
  < unit conversion >
    .ug2mg(Tum)     micron -> mg
    .mg2um(Tmg)     mg -> micron
    .mm2umGas(Tmm,Pa,degC)  gas thick. conv. mm -> micron
    .um2mmGas(Tum,Pa,degC)  gas thick. conv. micron -> mm
    .ThkStdGas(Pa,degC)     gas thick. conv. factor for this srWS-values
    .UidD()         LET unit names dict like; {Uid:'unit name',..}
    .Uid(Uid)       LET unit names
    .CnvF(Uid)      LET unit conversion factor used for;
                    LET(E,Uid) = _LET(E;Uid=0) * CnvF(Uid)
    .TrgCmpL()      Target material composition table; see ._TrgCmp
    < sheet info(2): min|max value of the internal tables >
    .{min|max}E     min|max of the E table
    .{min|max}LETe  min|max of the LET_elec table
    .{min|max}LETn  min|max of the LET_nucl table
    .{min|max}LETt  min|max of the LET_tot  table
    .{min|max}Rng   min|max of the Rng table
    .{min|max}StLng min|max of the Straggling Longitudinal table
    .{min|max}StLtr min|max of the Straggling Lateral table
    .maxLETe2E      E at max LET_elec
    .maxLETn2E      E at max LET_nucl
    .maxLETt2E      E at max LET_total
  < functions(1): E --> Value >
    .LETe(E) == .E2LETe(E)  dE/dX_elec.
    .LETn(E) == .E2LETn(E)  dE/dX_nucl.
    .LETt(E) == .E2LETt(E)  dE/dX_tot = elec+nucl
    .Rng(E)  == .E2Rng(E)   Range
    .StLng(E)== .E2StLng(E) Straggling Longitudinal
    .StLtr(E)== .E2StLtr(E) Straggling Lateral
  < functions(2): Value --> E ; inverse functions >
    .LETe2E(LETe,Uid,Hside) inv.func. of .LETe(E)
    .LETn2E(LETn,Uid,Hside) inv.func. of .LETn(E)
    .LETt2E(LETt,Uid,Hside) inv.func. of .LETt(E)
    .Rng2E(Rng)             inv.func. of .Rng(E)
    .StLng2E(StLng)         inv.func. of .StLng(E)
    .StLtr2E(StLtr)         inv.func. of .StLtr(E)
  < functions(3): Enew / Eold / related funcs. >
    .Enew(E,Tum)            new energy through this target; thick=Tum
    .EnewGas(E,Tmm,Pa,degC) new energy; for gas-target
    .Eold(E,Tum)            old energy before this target; thick=Tum
    .EoldGas(E,Tmm,Pa,degC) old energy; for gas-target
  < functions(4): work-sheet info >
    .info(modL,show)        info. of this srWS
    .plotLET(..)            plot ._LET{e|n|t} 
    .plotRngSt(..)          plot ._Range, ._StLng, ._StLtr 
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""    
    def _myclear(self):
        # < sheet-info(1) >
        self.wsName     = ""
        self.SRIMver    = ""
        self.IonZ       = 0
        self.IonA       = 0
        self.Beam       = ""
        self.Target     = ""
        self.TargetNm   = ""
        self.TrgDens    = 0.
        self.TrgDensAtm = 0.
        self.BraggC     = 0.
        self.Gas        = False
        self.GasPtbl    = 0.
        self.GasTtbl    = 0.
        self.Corded     = ""
        # < sheet info(2) >
        self.minE,     self.maxE     = 0., 0.
        self.minLETe,  self.maxLETe,  self.maxLETe2E = 0., 0., 0.
        self.minLETn,  self.maxLETn,  self.maxLETn2E = 0., 0., 0.
        self.minLETt,  self.maxLETt,  self.maxLETt2E = 0., 0., 0.
        self.minRng,   self.maxRng   = 0., 0.
        self.minStLng, self.maxStLng = 0., 0.
        self.minStLtr, self.maxStLtr = 0., 0.
        # < internal tables >
        self._E      = None     # ndarray
        self._LETe   = None
        self._LETn   = None
        self._LETt   = None
        self._LETeL, self._LETeH = None, None   # (E,LETe)
        self._LETnL, self._LETnH = None, None
        self._LETtL, self._LETtH = None, None
        self._Rng    = None
        self._StLng  = None
        self._StLtr  = None
        self._CnvF   = None
        self._TrgCmp = None     # [dict]
        # < fitting method >
        #self._intrp  = interpolate.interp1d     # scipy 線形補間
    
    def __init__(self):
        """  Constructor
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        self._myclear()

    """ DEBUG message header """
    def _msgH(self, fncNm:str) -> str:
        return( '#@ srWS[{}].{}:'.format(self.wsName,fncNm) )
    
    """ srWS : unit conversion
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def um2mg(self, Tum:float) -> float:
        """ Thickness convertion : micron -> mg
        Tum :   float   [um] thickness in micron
        <Ret>   float   [mg] = um * .TrgDens / 10
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( Tum * self.TrgDens / 10 )
        
    def mg2um(self, Tmg:float) -> float:
        """ Thickness convertion : mg -> micron
        Tmg :   float   [mg] thickness in mg
        <Ret>   float   [um] = mg * 10 / .TrgDens
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( Tmg * 10. / self.TrgDens )
    
    def _chkUid(self, Uid:int) -> bool:
        ok = True if(Uid in srimWS_Uid) else False
        if(not ok):
            msg = self._msgH('_chkUid')+'Err: Uid=[{}] invalid'.format(Uid)
            _Err(1,msg)
        return( ok )
    
    def UidD(self) -> dict:
        """ LET unit names dict
        <Ret> uidD  dict    { uid:unit, .. }
            uid:    int     LET unit ID
            unit:   str     LET unit
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( srimWS_Uid )

    def Uid(self, Uid:int) -> str:
        """ LET unit name
        <Ret>   str     LET unit, 'N/A' if Uid Err 
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(not self._chkUid(Uid)):
            return('N/A')
        return( srimWS_Uid[Uid] )
    
    def CnvF(self, Uid:int) -> float:
        """ LET unit conversion factor : LET*=CnvF
        Uid:    int     Unit ID 0..8
        <Ret>   float   CnvF, NaN if Uid Err
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._CnvF[Uid] if(self._chkUid(Uid)) else float('NaN'))

    def mm2umGas(self, Tmm:float, Pa:float, degC:float) -> float:
        """ Gas Thickness convertion : mm -> micron
        Tmm :   float   [mm] thickness in mm
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>   float   [um] = Tmm * 1000 * ThkStd(Pa,degC)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( Tmm * 1000. * self.ThkStdGas(Pa,degC) )
    
    def um2mmGas(self, Tum:float, Pa:float, degC:float) -> float:
        """ Gas Thickness convertion : mm -> micron
        Tum :   float   [um] thickness in micron
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>   float   [mm] = Tum / 1000 / ThkStd(Pa,degC)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( Tum / 1000. / self.ThkStdGas(Pa,degC) )

    def ThkStdGas(self, Pa:float, degC:float) -> float:
        """ Gas thick. conv. factor to "Standard P,T" of this SRIM table
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret> 
        CFct:   float   [1]     Conversion factor
                         == 1,  if this srWS is not Gas
                         == NaN,if (Pa < 0) | (degC < -273.15) 
        (usage) SRIMtbl_Thick[um] = Thick[um] * CFct
                Thick[um] = SRIMtbl_Thick[um] / CFct
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        msH = self._msgH('ThkStdGas')
        if(not self.Gas):
            msg= (msH + 'Err: [{}] is not Gas. '
            'Return conv. factor = 1').format(self.wsName)
            _Err(1,msg)
            return( 1. )
        if(Pa < 0.):
            msg= (msH + 'Err: Pa=[{}] < 0 Pa.'
            'Return conv. factor = 0').format(Pa)
            _Err(1,msg)
            return( float('NaN') )
        if(degC < -273.15):
            msg= (msH + 'Err: degC=[{}] < -273.15 C'
            'Return conv. factor = 0').format(Pa)
            _Err(1,msg)
            return( float('NaN') )
        P0 = self.GasPtbl
        T0 = self.GasTtbl
        return( (Pa / P0) * ((273.15 + T0) / (273.15 + degC)) )
        """
    note) in SRIM-2013                  Tst      Vst   Pst
    STP : Standard         Temp.& Pres.:  0 C(273.15K),22.7L,1013.25hPa(1atm)
    SATP: Standard Ambient Temp.& Pres.: 25 C(298.15K),24.8L,1000.00hPa(1bar)
    see) TRIM manual http://www.srim.org/SRIM/SRIM%2008.pdf
         8-12 Stopping Powers for Ions in Gases
    The stopping of ions in gases is very dependent on the gas pressure.
    STP (0 C, 760 mm). "TRIM always assumes a gas target is at STP."
    Pst*Vst/Tst = P*V/T = const         -> Vst  = V  *(P/Pst)*(Tst/T)
       Dens     = g / V                 -> Dst  = D  *(Pst/P)*(T/Tst)
       Tk [cm]  = [g/cm2] / Dens[g/cm3] -> Tkst = Tk *(P/Pst)*(Tst/T)
    But.. the [Compund Dictionary] sometimes seemed to be assumed
    different Temperature, Pressure and Density.
    Then, please check and describe the value of "Ptbl" and "Ttbl"
    in the SRIM data table by your own responsibility.
        """
    
    def TrgCmpL(self) -> list:
        """ Target material composition table
        <Ret>   [dict]  like;
        [{'AtomSym':'C','AtomNum':6,'Atom%':1.,'Mass%':1.},..]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._TrgCmp )
    
    """ srWS :  E --> Value
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def LETe(self, E:float, Uid:int=0) -> float:
        """  LETe(E) dE/dX electric part
        <Ret> .LETe(E) * CnvF(Uid) or 0 (if Uid error)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._intrpLET('LETe',E,Uid,self._LETe) )
    
    def LETn(self, E:float, Uid:int=0) -> float:
        """  LETn(E) dE/dX nuclear part
        <Ret> .LETn(E) * CnvF(Uid) or 0 (if Uid error)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._intrpLET('LETn',E,Uid,self._LETn) )
    
    def LETt(self, E:float, Uid:int=0) -> float:
        """  LETt(E) dE/dX total= nuclear + electric part
        <Ret> .LETt(E) * CnvF(Uid) or 0 (if Uid error)
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._intrpLET('LETt',E,Uid,self._LETt) )
     
    def _intrpLET(self, fncH, E, Uid, LET) -> float:
        if(   E < self.minE):
            v, ok = self.minE, False
        elif( E > self.maxE):
            v, ok = self.maxE, False
        else:
            v, ok = np.interp(E, self._E, LET), True            
        if(not ok):
            msg = (self._msgH(fncH)+'Warn: E=[{:.2e}] out of range.'
            '(min,max)=({:.2e},{:.2e})').format(E,self.minE,self.maxE)
            _Err(1,msg)
        return( v if(Uid==0) else v * self.CnvF(Uid) )
        
    def Rng(self, E:float) -> float:
        """ <Ret>  Rng(E) [um] Range """
        return( self._intrpTbl('Rng',E, self._Rng) )
    
    def StLng(self, E:float) -> float:
        """ <Ret>  StLng(E) [um] Straggling Longitudinal """
        return( self._intrpTbl('StLng',E, self._StLng) )
    
    def StLtr(self, E:float) -> float:
        """ <Ret>  StLng(E) [um] Straggling Lateral """
        return( self._intrpTbl('StLtr',E, self._StLtr) )
    
    def _intrpTbl(self, fncH, E, Vtbl) -> float:
        """  Vtbl[] --> E : interpolation 
        ._E[]          [MeV/A]
        Vtbl[]         [um]
        <Ret> Vtbl[E]  [um]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(   E < self.minE):
            v, ok = self.minE, False
        elif( E > self.maxE):
            v, ok = self.maxE, False
        else:
            v, ok = np.interp(E, self._E, Vtbl), True            
        if(not ok):
            msg = (self._msgH(fncH)+'Warn: E=[{:.2e}] out of range.'
            '(min,max)=({:.2e},{:.2e})').format(E,self.minE,self.maxE)
            _Err(1,msg)
        return( v )
        
    def E2LETe(self, E:float, Uid:int=0) -> float:
        """ == LETe(E,Uid) """
        return( self.LETe(E,Uid) )
    
    def E2LETn(self, E:float, Uid:int=0) -> float:
        """ == LETn(E,Uid) """
        return( self.LETn(E,Uid) )
    
    def E2LETt(self, E:float, Uid:int=0) -> float:
        """ == LETt(E,Uid) """
        return( self.LETt(E,Uid) )
    
    def E2Rng(self, E:float) -> float:
        """ == Rng(E) """
        return( self.Rng(E) )
    
    def E2StLng(self, E:float) -> float:
        """ == StLng(E) """
        return( self.StLng(E) )
    
    def E2StLtr(self, E:float) -> float:
        """ == StLtr(E) """
        return( self.StLtr(E) )
 
    """ srWS :  Value --> E ; inverse functions
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def LETe2E(self, LETe:float,*,
               Uid:int=0, Hside:bool=True) -> float:
        """  LETe --> E : inv.func. of LETe(E)
        LETe:   float   [Uid unit]
        Uid:    int     Unit ID#  (defa.=0)
        Hside:  bool    if(True) search at E-higher side of the LETe peak
                        else      ''       E-lower-side  ''
        <Ret> 
        E       float   [MeV/A]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._LET2E('LETe2E',LETe,Uid,Hside,
                    self._LETeL,self._LETeH,self.minLETe,self.maxLETe) )
    
    def LETn2E(self, LETn:float,*, Uid:int=0, Hside:bool=True) -> float:
        """  LETn --> E : inv.func. of LETn(E)
        see) LETe2E() for detail
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._LET2E('LETn2E',LETn,Uid,Hside,
                    self._LETnL,self._LETnH,self.minLETn,self.maxLETn) )
    
    def LETt2E(self, LETt:float,*, Uid:int=0, Hside:bool=True) -> float:
        """  LETt --> E : inv.func. of LETt(E)
        see) LETe2E() for detail
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        return( self._LET2E('LETt2E',LETt,Uid,Hside,
                    self._LETtL,self._LETtH,self.minLETt,self.maxLETt) )
    
    def _LET2E(self,fncH,let,Uid,Hside,letL,letH,letMin,letMax) -> float:
        """  .let{H|L} --> E : interpolation 
        .let{H|L}[0]    [MeV/A]
        .let{H|L}[1]    [MeV/(mg/cm2)]
        <Ret>  E        [MeV/A]
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        let = let if(Uid==0) else let / self.CnvF(Uid)
        if(   let < letMin):
            _let, ok = letMin, False
        elif( let > letMax):
            _let, ok = letMax, False
        else:
            _let, ok = let, True 
        if(not ok):
            msg = (self._msgH(fncH)+'Warn: let=[{:.2e}] out of range.'
            '(min,max)=({:.2e},{:.2e})').format(let,letMin,letMax)
            _Err(1,msg)
        if(Hside):  return( np.interp(_let, letH[1], letH[0]) )
        else:       return( np.interp(_let, letL[1], letL[0]) )
        """ note) scipy interpolate.interp1d() ->r#203)
            を試みたがヤメた。理由は、
            　LET(E) の逆関数は、多価関数になるので interp1dではダメ
            　また、shell effect のキンクなどもあるので
            BrgPk から検索方向を決めてsearchした方が単純明快である。、
            よって、srimWScnv 時に ._LETeL, .LETeH を作っておくことにした。
        """
    
    def Rng2E(self, Rng:float) -> float:
        """  Rng --> E : inv.func. of Rng(E)  """
        return( self._Tbl2E('Rng2E',Rng, self._Rng, self.minRng, self.maxRng) )
    
    def StLng2E(self, Str:float) -> float:
        """  StLng --> E : inv.func. of StLng(E)  """
        return( self._Tbl2E('StLng2E',Str, self._StLng, self.minStLng, self.maxStLng) )
    
    def StLtr2E(self, Str:float) -> float:
        """  StLtr --> E : inv.func. of StLtr(E)  """
        return( self._Tbl2E('StLtr2E',Str, self._StLtr, self.minStLtr, self.maxStLtr) )
    
    def _Tbl2E(self,fncH,val,Vtbl,Vmin,Vmax) -> float:
        if(   val < Vmin):
            _val, ok = Vmin, False
        elif( val > Vmax):
            _val, ok = Vmax, False
        else:
            _val, ok = val, True 
        if(not ok):
            msg = (self._msgH(fncH)+'Warn: val=[{:.2e}] out of range.'
                  '(min,max)=({:.2e},{:.2e})').format(val,Vmin,Vmax)
            _Err(1,msg)
        return( np.interp(_val, Vtbl, self._E) )

    """ srWS :  Enew / Eold / related funcs.
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def Enew(self, E0:float, Tum:float) -> float:
        """  E0 -> Thick[um] -> E1:new-energy
        E0:     float   [MeV/A] Beam energy
        Tum:    float   [um]    Target thickness 'in micron'
        <Ret>
        E1      float   [MeV/A] Beam new-energy through the Tum
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(self.Gas):
            msg= (self._msgH('Enew')+'Err: [{}] is Gas. '
            'Std.T,P = 25 C, 1013.25 hPa are assumed.').format(self.wsName)
            _Err(1,msg)
            return( self.EnewGas(E0,Tum/1000,101325.,25.) )
        return( self._Enew(E0,Tum) )
    
    def EnewGas(self, E0:float, Tmm:float, Pa:float, degC:float) -> float:
        """  E0 -> Thick[um],Pa,degC -> E1:new-energy : for Gas
        E0:     float   [MeV/A] Beam energy
        Tmm:    float   [mm]    Target thickness 'in mm'
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>
        E1      float   [MeV/A] Beam new-energy through the Tmm
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(not self.Gas):
            msg='#@ EnewGas:Err: [{}] is not Gas.'.format(self.wsName)
            _Err(2,msg)   
        Tum = self.mm2umGas(Tmm,Pa,degC)
        return( self._Enew(E0,Tum) )
    
    def Eold(self, E1:float, Tum:float) -> float:
        """  E0:old-energy <- Thick[um] <- E1
        E1:     float   [MeV/A] Beam energy
        Tum:    float   [um]    Target thickness 'in micron'
        <Ret>
        E0      float   [MeV/A] Beam old-energy before the Tum
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(self.Gas):
            msg= (self._msgH('Eold')+'Err: [{}] is Gas. '
            'Std.T,P = 25 C, 1013.25 hPa are assumed.').format(self.wsName)
            _Err(1,msg)
            return( self.EoldGas(E1,Tum/1000,101325.,25.) )
        return( self._Eold(E1,Tum) )
    
    def EoldGas(self, E1:float, Tmm:float, Pa:float, degC:float) -> float:
        """  E0:old-energy <- Thick[um],Pa,degC <- E1 : for Gas
        E0:     float   [MeV/A] Beam energy
        Tmm:    float   [mm]    Target thickness 'in mm'
        Pa:     float   [Pa]    Gas pressure
        degC:   float   [degC]  Gas temperature
        <Ret>
        E0      float   [MeV/A] Beam old-energy before the Tmm
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(not self.Gas):
            msg=self._msgH('EoldGas')+'Err: [{}] is not Gas.'.format(self.wsName)
            _Err(2,msg)   
        Tum = self.mm2umGas(Tmm,Pa,degC)
        return( self._Eold(E1,Tum) )
    
    def _Enew(self,E0,Tum) -> float:
        """  E0 -> Thick[um] -> E1:new-energy : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(Tum == 0.):  return( E0 )
        if(Tum < 0.):
            msg= self._msgH('_Enew')+'Warn: Tum[{}] < 0 '.format(Tum)
            _Err(1,msg)
            return( E0 )
        if(E0 == 0.):   return( 0. )
        if(E0 <= 0.):   return( 0. )
        R0 = self.Rng(E0)
        R1 = R0 - Tum
        if(R1 <=0 ):    return( 0. )
        E1 = self.Rng2E( R1 )
        return( E1 )
    
    def _Eold(self,E1,Tum) -> float:
        """  E0:old-energy <- Thick[um] <- E1 : common
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(Tum == 0.):  return( E1 )
        if(E1 <= 0.):   return( 0. )
        R1 = self.Rng(E1)
        R0 = R1 + Tum
        E0 = self.Rng2E( R0 )
        return( E0 )
    
    """ srWS : info.
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""

    def info(self, modL:list=[0], *, 
             intrvE:int=1, show:bool=True ) -> list:
        """ Info. of this srimWS
        <optional>  see) srimWSdump()
        modL:   [int,..]    dump mode list
            0: simple info. ; wsName,Beam,Target
            1: other  info. ; TrgDens, BraggC, SRIMver, Corded
            2: target compo.; TrgCmp table
            3: conv. factor ; LET unit conversion table
            4: table size   ; {min|max} of E, LETe,n,t, Rng, StLng, StLtr
            5: table contens; E, LET{e|n|t}, Rng, StLng, StLtr
            6: table size   ; LET{e|n|t}{L|H}
           -1: all of above 
        <optional: kwd args> 
        intrvE: int         when mod=5, skip E interval 
        show:   bool        show dump result as print()
        <Ret>   
        nothing             if show=True
        txL:    [str,..]    else dump result as list of text    
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        txL = srimWSdump( self, modL=modL, intrvE=intrvE, show=show)
        if(not show):
            return( txL )

    def plotLET(self, *, 
                LETe:bool=True, LETn:bool=True, LETt:bool=True,
                rngX:tuple=(1e-5,1e3), rngY=(1e-2,1e2)) -> tuple:
        """ Plot ._LET{e|n|t} 
        <optional: kwd args>
        LET{e|n|t}:     bool    plot SW for LET{e|n|t}
        rngX,rngY:      tuple   plot range (min,max)
        <Ret> fg,ax             figure, axies
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        fg,ax = srimWSplot_LET(self, LETe=LETe, LETn=LETn, LETt=LETt,
                           rngX=rngX, rngY=rngY)
        return(fg, ax)
    
    def plotRngSt(self, *, 
                  Rng:bool=True, StLng:bool=True, StLtr:bool=True,
                  rngX:tuple=(1e-5,1e3), rngY=(1e-3,1e6) ) -> tuple:
        """ Plot ._Range, ._StLng, ._StLtr
        <optional: kwd args>
        Rng,StLng,StLtr: bool   plot SW for Rng,StLng,StLtr
        rngX,rngY:  tuple       plot range (min,max)
        <Ret> fg,ax             figure, axies
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        fg,ax = srimWSplot_RngSt(self, Rng=Rng, StLng=StLng, StLtr=StLtr,
                            rngX=rngX, rngY=rngY )
        return(fg, ax)

    """ srWS : for print()
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    
    def __str__(self):
        tx = self.wsName + ': '
        tx += atomSym().Asym(self.IonA,self.IonZ) + ' -> '
        tx += self.Target + ' : ' + self.Corded
        return(tx)


""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':
    print('TEST : srimWSname_split()') 
    print(srimWSname_split('srim84Kr_Air_NTP'))
    srWS = srimWS()
    print('@@main: -- Fin ---')


""" ========================================================
        ref) Numbered ref
    ========================================================
    
// 内挿補間 //
r#201) Qiita2018 様々な補間法と最小2乗法をPythonで理解する
    https://qiita.com/maskot1977/items/2a55c087974d0ac5e843
  Qiita2019 Scipy.interpolate を使った様々な補間法
    https://qiita.com/maskot1977/items/913ef108ff1e2ba5b63f
  Qiita2019 Numpy.polyfit を使ったカーブフィッティング
    https://qiita.com/maskot1977/items/f9e84fb6e62f7015d6c4
  Qiita2018 数値補間手法の食べ比べ
    https://qiita.com/sus304/items/1eeb22d4456c2fb1717d

r#202)　Numpy interp
    https://numpy.org/doc/stable/reference/generated/numpy.interp.html
〇  numpy.interp 使い方
    https://runebook.dev/ja/docs/numpy/reference/generated/numpy.interp
    
r#203) SciPy interpolation
    https://docs.scipy.org/doc/scipy/reference/interpolate.html
〇  Qiita2021 補間 scipy.interpolate.interp1d おさらい
    https://qiita.com/shiba54/items/0511d3ba978909590b3d
      f = interp1d(xdat,ydat,assumed_sorted=False)
      x が昇順でない場合に自動的に x, y の組み合わせが x で昇順ソート
      y = f(x) 
        
    ========================================================
        ref) Others
    ======================================================== 
    nkmk Pythonで文字列を分割（区切り文字、改行、正規表現、文字数） 
    https://note.nkmk.me/python-split-rsplit-splitlines-re/
r#901) 第二引数maxsplitで最大分割回数を指定
r#902) 正規表現にマッチした部分で分割

r#903) nkmk Pythonの正規表現モジュールreの使い方（match, search, subなど）
    https://note.nkmk.me/python-re-match-search-findall-etc/
------------------------------------------------------------------
"""