# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   srimWScnv.py
    Descr.: def. class srimWScnv ; Convert srWS <-> ExcelWS
    Update: 220403 build start Ay
            220523 v.1.0 done 
    Last:   230318 v.1.1 refactored
    
@author: Ayoshida(RIKEN,Nishina center)
"""

""" import: standard lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
import  numpy   as  np
import  re

""" import: my custom lib.       
----.---.---.---.---.---.---.---.---.---.---.---.---"""
#import  aypylib             as ay 

""" import: from this package   
----.---.---.---.---.---.---.---.---.---.---.---.---"""
from    pySRIMfit   import  (_Err,srimWS,srimWS_TrgCmp,srimWSname_AZT)

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
        class srimWScnv
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """
class srimWScnv( object ):
    """ SRIM Work-Sheet Converter
Params. for constructor:
    non
Attributes:
    .srWS     srimWS
    .xlWS     xlwings.sheet
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    """ Const.Def.      Work Sheet Structure
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    """ Work Sheet format : cell address """
    # < sheet-info(1) >
    _celVer   = 'C3'    # SRIM ver
    _celIonZ  = 'C4'    # Ion Z
    _celIonA  = 'C5'    # Ion A
    _celTgNS  = 'C6'    # Target name Short
    _celTgNL  = 'C7'    # Target name Long
    _celTgDG  = 'C8'    # Target density in [g/cm3]
    _celTgDA  = 'C9'    # Target density in [atoms/cm3]
    _celBrgC  = 'C10'   # Bragg Correction [%]
    _celEmin  = 'C12'   # column # of Emin
    _clmEmin  =  20     # column # at Etbl-top 
    _celEmax  = 'C13'   # column # of Emax
    _celGas   = 'F3'    # Target is Gas ?
    _celGasPtbl='C14'   # if Gas, P[Pa]   asuumed in this SRIM data table
    _celGasTtbl='C15'   # if Gas, T[degC] asuumed in this SRIM data table
    _celCorded= 'P6'    # Corded info
    # < internal tables >
    _tblE     = 'D20'   # E[MeV/A] table
    _tblLETe  = 'E20'   # dE/dX_elec [MeV/(mg/cm2)]
    _tblLETn  = 'F20'   # dE/dX_nucl [MeV/(mg/cm2)]
    _tblLETt  = 'G20'   # dE/dX_tot  [MeV/(mg/cm2)]
    _tblRng   = 'J20'   # Range[um]
    _tblStLng = 'M20'   # Straggling Longitudinal[um]
    _tblStLtr = 'P20'   # Straggling Lateral[um]
    _tblCnvF  = 'K6'    # LET unit conversion factors
    _tblTrgCmp= 'F6'    # arget material compositions

    """ class Attributes
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""    
    def _myclear(self):
        pass
        
    def __init__(self):
        """  Constructor
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        self._myclear()
    
    def xlWS2srWS(self, xws )-> object:
        """ Excel WkSht -> pySRIM WkSht conversion
        xws:  xlwings.sheet     input Excel work sheet (xlwings)
        <Ret> 
        sws:  srimWS            converted output
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        def _rd_tbl(xws,cel0,n) -> object:
            """  Read column shape table
            xws:    input Excel work sheet (xlwings)
            cel0:   left-up cell addr for the table
            n:      num. of rows
            <Ret>   np.array
            ----.---.---.---.---.---.---.---.---.---.---.---.---"""
            rng0 = xws.range(cel0)
            rng1 = rng0.offset(n,0)
            vals = xws.range(rng0,rng1).value
            return( np.array( vals, dtype=float ) )
        
        def _rd_tbl_TrgCmp(xws) -> list:
            """  Read TrgCmp table
            xws:    input Excel work sheet (xlwings)
            <Ret>   [dict] of TrgCmp
            ----.---.---.---.---.---.---.---.---.---.---.---.---"""
            rng0 = xws.range(self._tblTrgCmp)
            dctL = []
            for clm in range(8):
                rng1 = rng0.offset(clm,0)
                rng2 = rng1.offset(0,3)
                vals = xws.range(rng1,rng2).value
                if(vals[0] is None): break
                vals[1] = int(vals[1])
                dctL.append( dict(zip(srimWS_TrgCmp,vals)) )
            if(len(dctL) < 1):
                _Err(1,'#@ srimWScnv: TrgCmp is blank')
            return( dctL )
        
        def _minmax(E, V) -> tuple:
            """  Find min/max of E vs V tables
            E:          ndary   Energy table
            V:          ndary   Value  table (like LETe）
            <Ret> (minV,minE,maxV,maxE)
            minV,maxV     float   min/max of Value
            minE,maxE   float   E at min/max of Value
            ----.---.---.---.---.---.---.---.---.---.---.---.---"""
            minV, maxV = V.min(), V.max()
            #maxE = E[ np.where(V == maxV) ]  ダメ ret=[] になる
            minE = E[ V.argmin() ]
            maxE = E[ V.argmax() ]
            return( (minV,minE,maxV,maxE))
        
        def _sepLH(E, V) -> tuple:
            """  separate (E[] vs V[]) at Lower/Higher-side of its V-peak
            E:      ndary   Energy table
            V:      ndary   Value  table (like LETe）
            <Ret> ((EL,VL),(EH,VH))
            E{L|H}  ndary   E Lower/Higher-side
            V{L|H}  ndary   corresponding V Lower/Higher-side
            ----.---.---.---.---.---.---.---.---.---.---.---.---"""
            ixPk = V.argmax()
            EL,EH = E[0:ixPk+1], E[ixPk:]
            VL,VH = V[0:ixPk+1], V[ixPk:]
            EH,VH = EH[::-1], VH[::-1]  # inverse-sort to use np.interp()
            return( (EL,VL),(EH,VH) )
    
        #print('converting =',xws.name)
        sws = srimWS()
        sws.wsName     = xws.name
        # read cells
        sws.SRIMver    = xws.range(self._celVer).value
        sws.IonZ       = int(xws.range(self._celIonZ).value)
        sws.IonA       = int(xws.range(self._celIonA).value)
        sws.Target     = xws.range(self._celTgNS).value
        sws.TargetNm   = xws.range(self._celTgNL).value
        if(sws.Target   is None): 
            _Err(1,'#@ srimWScnv: Target is blank')
            sws.Target   = ''
        if(sws.TargetNm is None):
            _Err(1,'#@ srimWScnv: TargetNm is blank')
            sws.TargetNm = ''
        sws.TrgDens    = xws.range(self._celTgDG).value
        sws.TrgDensAtm = xws.range(self._celTgDA).value
        sws.BraggC     = xws.range(self._celBrgC).value  # read as [%]
        gas            = xws.range(self._celGas).value
        sws.Gas = True if(gas == 'Gas') else False
        if(sws.Gas):
            sws.GasPtbl= xws.range(self._celGasPtbl).value
            sws.GasTtbl= xws.range(self._celGasTtbl).value
        sws.Corded     = xws.range(self._celCorded).value
        if(sws.Corded is None):
            sws.Corded = ''
        # read tables
        _clmEmin   = int(xws.range(self._celEmin).value)
        _clmEmax   = int(xws.range(self._celEmax).value)
        n = _clmEmax - _clmEmin
        sws._E         = _rd_tbl(xws,self._tblE,    n)
        # check E tbl is monotonically increase
        if(not np.all(np.diff(sws._E) > 0)):
            msg= '#@ srimWScnv: E tbl is NOT monotonically increased'
            raise ValueError(msg)
        sws._LETe      = _rd_tbl(xws,self._tblLETe, n)
        sws._LETn      = _rd_tbl(xws,self._tblLETn, n)
        sws._LETt      = _rd_tbl(xws,self._tblLETt, n)
        sws._Rng       = _rd_tbl(xws,self._tblRng,  n)
        sws._StLng     = _rd_tbl(xws,self._tblStLng,n)
        sws._StLtr     = _rd_tbl(xws,self._tblStLtr,n)
        nda            = _rd_tbl(xws,self._tblCnvF,7)
        sws._CnvF      = np.hstack([nda[4],nda])  # Uid=0 == Uid=5
        #print(nda,sws._CnvF)
        sws._TrgCmp    = _rd_tbl_TrgCmp(xws)
        # find min/max
        sws.minE,    sws.maxE     = sws._E[0], sws._E[-1]
        sws.minLETe,_,sws.maxLETe,sws.maxLETe2E= _minmax(sws._E,sws._LETe)
        sws.minLETn,_,sws.maxLETn,sws.maxLETn2E= _minmax(sws._E,sws._LETn)
        sws.minLETt,_,sws.maxLETt,sws.maxLETt2E= _minmax(sws._E,sws._LETt)
        sws.minRng,  _,sws.maxRng,_   = _minmax(sws._E,sws._Rng)
        sws.minStLng,_,sws.maxStLng,_ = _minmax(sws._E,sws._StLng)
        sws.minStLtr,_,sws.maxStLtr,_ = _minmax(sws._E,sws._StLtr)
        # E vs LET at Lower/Higher-side of its peak
        sws._LETeL,sws._LETeH = _sepLH(sws._E,sws._LETe)
        sws._LETnL,sws._LETnH = _sepLH(sws._E,sws._LETn)
        sws._LETtL,sws._LETtH = _sepLH(sws._E,sws._LETt)
        return(sws)
    
    def srWS2xlWS(self, sws, xws)-> object:
        """ pySRIM WkSht -> Excel WkSht conversion
        sws:    srimWS        input srimWS
        xws:    xlwings.sheet prototype sheet to write
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        def _wr_tbl_TrgCmp(xws,dctL):
            """  Write TrgCmp table
            xws:    input Excel work sheet (xlwings)
            dctL:   [dict] of TrgCmp
            ----.---.---.---.---.---.---.---.---.---.---.---.---"""
            rng0 = xws.range(self._tblTrgCmp)
            for clm,dct in enumerate(dctL):
                if(clm >= 7): break
                rng1 = rng0.offset(clm,0)
                xws.range(rng1).value = list(dct.values())

        #print('converting =',sws.wsName)
        xws.range(self._celVer).value   = sws.SRIMver
        xws.range(self._celIonZ).value  = sws.IonZ
        xws.range(self._celIonA).value  = sws.IonA
        xws.range(self._celTgNS).value  = sws.Target
        xws.range(self._celTgNL).value  = sws.TargetNm
        xws.range(self._celTgDG).value  = sws.TrgDens
        xws.range(self._celTgDA).value  = sws.TrgDensAtm
        xws.range(self._celBrgC).value  = sws.BraggC
        if(sws.Gas):
            xws.range(self._celGas).value       = 'Gas'
            xws.range(self._celGasPtbl).value   = sws.GasPtbl
            xws.range(self._celGasTtbl).value   = sws.GasTtbl
        xws.range(self._celCorded).value = sws.Corded
        # write tables
        n = len(sws._E)
        xws.range(self._celEmin).value  = self._clmEmin
        xws.range(self._celEmax).value  = self._clmEmin + n
        #self._wr_tbl(xws,self._tblLETe,n,sws._E)
        xws.range(self._tblE    ).options(transpose=True).value = sws._E
        xws.range(self._tblLETe ).options(transpose=True).value = sws._LETe
        xws.range(self._tblLETn ).options(transpose=True).value = sws._LETn
        xws.range(self._tblLETt ).options(transpose=True).value = sws._LETt
        xws.range(self._tblRng  ).options(transpose=True).value = sws._Rng
        xws.range(self._tblStLng).options(transpose=True).value = sws._StLng
        xws.range(self._tblStLtr).options(transpose=True).value = sws._StLtr
        xws.range(self._tblCnvF ).options(transpose=True).value = sws._CnvF
        _wr_tbl_TrgCmp(xws,sws._TrgCmp)
        return( xws )

    def srOut2xlWS(self, fnOut:str, xws)-> object:
        """ SRIM.out -> Excel WkSht conversion
        fnOut   str             SRIM.output text file path
        xws:    xlwings.sheet   prototype sheet to write
        <Ret> 
        xws:    xlwings.sheet   converted output
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        def _skip2(kwd):
            nonlocal iln
            for i in range(iln, eln):
                if(kwd in lnsT[i]): break
            iln = i
            return(lnsT[i],lnsC[i])
        
        #print('converting fn=',fnOut)
        with open(fnOut) as f:
            lnsT = f.readlines()                # lines as Text 
        lnsC = [ln.split() for ln in lnsT]      # lines as CSV (space separated)
        iln, eln = 0, len(lnsT)                 # line ptr, end ptr

        lT,lC = _skip2('SRIM version')
        xws.range(self._celVer).value   = lC[3]
        lT,lC = _skip2('Disk File Name =')
        Trg = re.search("(?<=in).*?(?=.txt)",lT).group()    # ->r#811)
        Trg = Trg.lstrip()
        xws.range(self._celTgNS).value  = ''
        xws.range(self._celTgNL).value  = Trg
        lT,lC = _skip2('Ion =')
        IonZ = int( re.search("(?<=\[).+?(?=\])",lC[3]).group() ) # ->r#812)
        IonA = int( lC[7] )
        xws.range(self._celIonZ).value  = IonZ
        xws.range(self._celIonA).value  = IonA
        lT,lC = _skip2('Target Density =')
        xws.range(self._celTgDG).value  = lC[3]
        xws.range(self._celTgDA).value  = lC[6]
        if('Target is a GAS' in lnsT[iln+1]):
            xws.range(self._celGas).value       = 'Gas'
            xws.range(self._celGasPtbl).value   = 101325.  # [Pa]
            xws.range(self._celGasTtbl).value   = 20.
        lT,lC = _skip2('Percent')
        iln += 2    #skip 2 lines
        rng0 = xws.range(self._tblTrgCmp)
        for clm in range(7):
            vals = lnsT[iln+clm].split()
            if(len(vals) != 4): break
            rng1 = rng0.offset(clm,0)
            xws.range(rng1).value = vals
        lT,lC = _skip2('Bragg Correction =')
        xws.range(self._celBrgC).value  = lC[3]
        if(not 'Stopping Units =  MeV / (mg/cm2)' in lnsT[iln+1]):
            print('err:srOut2xlWS Units !=  MeV / (mg/cm2)')
            return( xws )
        lT,lC = _skip2('Straggling')
        iln += 2    #skip 2 lines
        rng0 = xws.range(self._tblE)
        for clm in range(eln-iln):
            v = lnsT[iln+clm].split()
            print('\r Etbl[{}]'.format(clm),end='')
            if(len(v) != 10): break
            E,EF       = self._cnvE(v[1],float(v[0]),IonA)
            Rng,RngF   = self._cnvR(v[5],float(v[4]))
            SLng,SLngF = self._cnvR(v[7],float(v[6]))
            SLtr,SLtrF = self._cnvR(v[9],float(v[8]))
            vl = [v[0],v[1],E,v[2],v[3],float(v[2])+float(v[3]),\
                  v[4],v[5],Rng,v[6],v[7],SLng,v[8],v[9],SLtr]
            xws.range(rng0.offset(clm,-2)).value = vl
            xws.range(rng0.offset(clm, 0)).number_format = EF
            xws.range(rng0.offset(clm, 6)).number_format = RngF
            xws.range(rng0.offset(clm, 9)).number_format = SLngF
            xws.range(rng0.offset(clm,12)).number_format = SLtrF
        print()
        xws.range(self._celEmax).value  = self._clmEmin+clm-1
        lT,lC = _skip2('Multiply')
        iln += 2    #skip 2 lines
        rng0 = xws.range(self._tblCnvF)
        for clm in range(8):
            vals = lnsT[iln+clm].split()
            if(len(vals) < 2): break
            xws.range(rng0.offset(clm,0)).value = vals[ 0]
        nm = srimWSname_AZT(IonA,IonZ,Trg)
        #print('WSname=',nm)
        xws.name = nm[0:15]     # WkSht.name < 31 chars     
        return( xws )

    def srCSV2xlWS(self, xwsc, xws)-> object:
        """ SRIM.out CSV WkBook -> Excel WkSht conversion
        xwsc    xlwings.sheet   SRIM.output CSV WkSht
        xws:    xlwings.sheet   prototype sheet to write
        <Ret> 
        xws:    xlwings.sheet   converted output
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""        
        def _skip2(kwd):
            nonlocal iln
            for i in range(iln, eln):
                if(kwd in lnsC[i]): break
            iln = i
            return(lnsC[i])
        
        #print('converting =',xwsc.name)
        rng  = xwsc.used_range      # range of used cells
        lnsC = rng.value            # read cell-values as list
        iln, eln = 0, len(lnsC)
        
        lC = _skip2('SRIM')
        xws.range(self._celVer).value   = lC[4]
        lC = _skip2('Disk')
        Trg = lC[8][0:16]
        xws.range(self._celTgNS).value  = ''
        xws.range(self._celTgNL).value  = Trg
        lC = _skip2('Ion')
        IonZ = int( re.search("(?<=\[).+?(?=\])",lC[4]).group() ) # ->r#812)
        IonA = int( lC[8] )
        xws.range(self._celIonZ).value  = IonZ
        xws.range(self._celIonA).value  = IonA
        lC = _skip2('Target')
        xws.range(self._celTgDG).value  = lC[4]
        xws.range(self._celTgDA).value  = lC[7]
        if('GAS' in lnsC[iln+1]):
            xws.range(self._celGas).value       = 'Gas'
            xws.range(self._celGasPtbl).value   = 101325.  # [Pa]
            xws.range(self._celGasTtbl).value   = 20.
        lC = _skip2('Percent')
        iln += 2    #skip 2 lines
        rng0 = xws.range(self._tblTrgCmp)
        for clm in range(7):
            if(lnsC[iln+clm][2] is None): break
            rng1 = rng0.offset(clm,0)
            xws.range(rng1).value = lnsC[iln+clm][1:4]
        lC = _skip2('Bragg')
        xws.range(self._celBrgC).value  = lC[4]
        #if(not 'Stopping Units =  MeV / (mg/cm2)' in lnsT[iln+1]):
        #    print('err:srOut2xlWS Units !=  MeV / (mg/cm2)')
        #    return( xws )
        lC = _skip2('Straggling')
        iln += 2    #skip 2 lines
        rng0 = xws.range(self._tblE)
        for clm in range(eln-iln):
            print('\r Etbl[{}]'.format(clm),end='')
            v = lnsC[iln+clm]
            if(v[2] is None): break
            E,EF       = self._cnvE(v[ 2],float(v[ 1]),IonA)
            Rng,RngF   = self._cnvR(v[ 6],float(v[ 5]))
            SLng,SLngF = self._cnvR(v[ 8],float(v[ 7]))
            SLtr,SLtrF = self._cnvR(v[10],float(v[ 9]))
            vl = [v[1],v[2],E,v[3],v[4],float(v[3])+float(v[4]),\
                  v[5],v[6],Rng,v[7],v[8],SLng,v[9],v[10],SLtr]
            xws.range(rng0.offset(clm,-2)).value = vl
            xws.range(rng0.offset(clm, 0)).number_format = EF
            xws.range(rng0.offset(clm, 6)).number_format = RngF
            xws.range(rng0.offset(clm, 9)).number_format = SLngF
            xws.range(rng0.offset(clm,12)).number_format = SLtrF
        print()
        xws.range(self._celEmax).value  = self._clmEmin+clm-1
        lC = _skip2('Multiply')
        iln += 2    #skip 2 lines
        rng0 = xws.range(self._tblCnvF)
        for clm in range(8):
            if(lnsC[iln+clm][2] is None): break
            v = lnsC[iln+clm]
            xws.range(rng0.offset(clm,0)).value = v[1]
        nm = xwsc.name
        #print('WSname=',nm)
        xws.name = nm 
        return( xws )

    def _cnvE(self,unit:str,E:float,A:int) -> tuple:
        """  conv. E [unit] --> E [MeV/A]
        unit:   str     unit for E
        E:      float   E [unit]
        A:      int     IonA
        ret.)
        E:      float   E [MeV/A]
        fmt:    str     xl cell format
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(unit == 'eV'):   E = E / 1000000. / A
        if(unit == 'keV'):  E = E / 1000. / A
        if(unit == 'MeV'):  E = E / A
        if(unit == 'GeV'):  E = E * 1000. / A
        fmt = '0.0000'
        if(E >= 1000.):      fmt = '0.00E+00'
        if(E <= 0.001):      fmt = '0.00E-00'
        return( E,fmt )

    def _cnvR(self,unit,R):
        """  conv. Rng, StLon, StLtr [unit] --> Val [um]
        unit:   str     unit for Rng, StLon, StLtr
        R:      float   R [unit]
        ret.)
        R:      float   R [um]
        fmt:    str     xl cell format
        ----.---.---.---.---.---.---.---.---.---.---.---.---"""
        if(unit == 'A'):    R = R / 1000. / 10.
        if(unit == 'mm'):   R = R * 1000.
        if(unit == 'm'):    R = R * 1000000.
        if(unit == 'km'):   R = R * 1000000000.
        fmt = '0.000'
        if(R >= 1000.):      fmt = '0.00E+00'
        if(R <= 0.001):      fmt = '0.00E-00'
        return( R,fmt )


""" ========================================================
            Test Main
    ========================================================"""
if __name__ == '__main__':
    print('PLZ. use _test/t_srimWBcnv_*.__main__')
    print('@@main: -- Fin ---')
