# -*- coding: utf-8 -*-
"""
    Module: pySRIMfit
    File:   t_pySRIMfit.py
    Descr.: TEST pySRIMfit module
    Update: 220515 build start Ay
            220522 v.1.0 done 
    Last:   230310 v.1.1 refactored

@author: Ayoshida(RIKEN,Nishina center)
"""

""" .===.===.===.===.===.===.===.===.===.===.===.===.===.===.===.===
            t_pySRIMfit.py の 使い方
        ============================
[0] はじめに
============================================
  ここでは、pySRIMfit の 簡単な使い方と、module 構造 を説明します。

[1] 簡単な使い方
============================================
    import  pySRIMfit   as  srim
    sr= srim.mySRIM()           <-- class mySRIM() インスタンス生成
    sr.info()　                  <-- シート名一覧表を表示 
    sr.maxLETt('srim40Ar_Si')   <-- 指定したシート名の maxLETt 値を表示
　※ 今までの Excel版 SRIMfit と同じ関数名が使えます。
    詳しくは、class srimWS(), srimWB(), e5SRIM() を見て下さい。 

[2] module 構造
============================================
DOS> tree /F ~/pyE5A
│  t_pySRIMfit.py       <-- "This file"
│  upd_MySRIMwb.py      <-- MySRIMwb アップデーター
├─pySRIMfit
│  │  srimWS.py             <-- class srimWS()
│  │  srimWB.py             <-- class srimWB()
│  │  e5SRIM.py             <-- class e5SRIM()
│  │  mySRIM.py             <-- class mySRIM( srimWB, e5SRIM )
│  │  srimMisc.py           <-- misc. common func.
│  │  srimWScnv.py          <-- class srimWScnv()
│  │  srimWBcnv.py          <-- srimWBcnv_*  func. 
│  │  srimWBupdater.py      <-- MySRIMwb.pickle updater
│  ├─_MySRIMwb
│  │      MySRIMwb.pickle       <-- pySRIMfit DB
│  │      MySRIMwb.xlsx         <-- the DB saved as .xlsx
│  ├─_proto
│  │      _MySRIMwb_prtyp.xlsx  <-- MySRIMwb.xlsx WkSheet proto-type
│  ├─_test
│  │  │  t_srimWBcnv_csv2xl.py      <-- misc. test scripts
│  │  │  ~ ~ ~
│  │  ├─add_SRIMoutCSV_eg           <-- misc. data for test scripts
│  │  │      SRIMout_084Kr.xlsx
│  │  ├─add_SRIMoutTxt_eg
│  │  │      Argon in Air, Dry (ICRU-104) (gas).txt
│  │  ├─_MySRIMwb_eg
│  │  |   └─_samples                <-- misc. sample-data for test scripts
│  │  ├─xw_test                     <-- pySRIMfit wapper for xlwings 
│  │  │      xw_pySRIMfit.py            (under construction)
│  │  │      XW_test.xlsm

　　・ class srimWS() が、mySRIMwb.xlsx の　各シートに相当します。
      各シートの数値:Attribution を参照する method が定義されています。
    class srimWB() が、mySRIMwb.xlsx の　全シートに相当します。
    　シートを検索する機能、複数シートを同時に用いた複合関数などの method 定義
　　・ class e5SRIM() が、E5Aコース専用の method 定義
  ・ class mySRIM() が、srimWB() と e5SRIM（) の両方の method を継承し、
　　　__init()__ の際に、MySRIMwb.pickle を自動open します。
   その際、multiple-open にならない様に制御しています。
  
　　・ 一般 user は、
     class mySRIM() を使って下さい。
　　・ 自分用に開発したい user は、
      class mySRIM()　に、独自開発した method class を継承させて下さい　

以上です。
====.===.===.===.===.===.===.===.===.===.===.===.===.===.===.=== === """

""" ========================================================
            Main
    ========================================================"""
if __name__ == '__main__':

    """ Test-CTRL Sw ---.---.---.---.---.---.---.---.---"""
    SwTest01 = True     # very simple examples
    SwTest02 = False    # make sr.Rng() table
    SwTest03 = False    # E5A beam-line: min/max LET
    SwTest04 = False    # plot LET, Range, Straggling 
    SwTest05 = False    # info srimWS, srimWB 
    
    import  pySRIMfit   as  srim
    sr= srim.mySRIM()   # load MySRIMwb.pickle
    sr.info()
    
    """ Test01 : HowToUse ; very simple examples
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest01):
        print('====== Test-01 ======')
        wsn = 'srim40Ar_Si'     # work sheet name
        print("sr.maxLETt('{}')= {:.2f} [{}] ; using sr.func()".format(
                   wsn, sr.maxLETt(wsn), sr.LETunit(0) ) ) 
        srWs = sr.getWS(wsn)    # srimWS object
        print("ws.maxLETt()= {:.2f} ; using srimWS object".format(
            srWs.maxLETt ) )  

    """ Test02 : make sr.Rng() table
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest02):
        print('====== Test-02 ======')
        wsnL=['srim12C_Si','srim20Ne_Si','srim40Ar_Si',
              'srim56Fe_Si','srim84Kr_Si']
        EL=[70,90,95,100,110,120,135,140,150,160]
        print('       Beam E =',end='')
        for E in EL:
            print('{:5.0f}'.format(E),end='')
        print(' [MeV/A]')
        for wsn in wsnL:
            print('{:<14}:'.format(wsn),end='')
            for E in EL:
                print('{:5.1f}'.format(sr.Rng(wsn,E)/1000),end='')
            print(' [mm]') 

    """ Test03 : E5A beam-line: min/max LET
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest03):
        print('====== Test-03 ======')
        Beam, BeamE = '84Kr', 69.59     # [MeV/A]
        Pa, degC =  1010.*100., 23.7    # [Pa],[C] Air press. & temp.
        #        Material  thick [um or mm]
        Matr=[  ['Au',      45.8 ],     # Beam Scattering foil 
                ['Kapton',  51.2 ],     # Vacuum Foil
                ['Al',       0.60],     # IC1;Ion Chamber Alumi Foil
                ['Mylar',   14.54],     # IC1 Mylar Foil
                ['Mylar',   10.2 ],     # PL1;Plastic Scintilator Mylar Foil
                ['Al',       0.4 ],     # PL1 Alumi on Mylar
                ['EJ212',  100.0 ],     # PL1 Scinti (EJ-212)
                ['Air',    145.0 ],     # [mm] Air1 gap; Kapton ~ EdegOut 
                ['Air',    200.0 ]      # [mm] Air2 gap; EdegOut ~ Sample
            ]
        #       Sample Mold [um]
        Smpl=[  ['Al',       0.0 ],     # mold#1
                ['Epoxy',  150.0 ],     # mold#2
                ['Si',       5.0 ],     # Device Layer#1
                ['Si',      10.0 ]      # Device Layer#2
            ]
        Targ = Matr + Smpl
        E, wsnSi = BeamE, 'srim'+Beam+'_Si'
        print(' #, Matr    ,  Thick,  Eaftr    LET, Range')
        print('            ,[um|mm],  [MeV/A],    , [um] ')
        print('    in Vacc ,         {:6.2f}, {:5.2f}, {:6.1f}'.format( 
            E, sr.LETt(wsnSi,E), sr.Rng(wsnSi,E)))
        for i,trg in enumerate(Targ):
            wsn = 'srim'+Beam+'_'+trg[0]
            if(sr.Gas(wsn)):
                E = sr.EnewGas(wsn,E,trg[1],Pa,degC)
            else:                
                E = sr.Enew(wsn,E,trg[1])
            LET = sr.LETt(wsnSi,E)
            Rng = sr.Rng(wsnSi,E)
            msg = '{:2d}, {:<8}, {:6.2f}, {:6.2f}, {:5.2f}, {:6.1f}'.format(
                i,trg[0],trg[1],E,LET,Rng)
            print(msg)

    """ Test04 : plot LET, Range, Straggling 
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest04):
        print('====== Test-04 ======')
        wsn = 'srim40Ar_Si'
        sr.WSplot_LET(wsn,LETe=True,LETn=True,LETt=True, 
            rngX=(1e-5,1e3), rngY=(1e-2,1e2))
        sr.WSplot_RngSt(wsn,Rng=True,StLng=True,StLtr=True, 
            rngX=(1e-5,1e3), rngY=(1e-3,1e6))
        wsnL = ['srim12C_Si','srim40Ar_Si','srim84Kr_Si','srim129Xe_Si']
        sr.WSplot_LETcmp(wsnL,LETe=False,LETn=False,LETt=True, 
            rngX=(1e-2,1e3), rngY=(1e-2,1e2))
        sr.WSplot_RngStcmp(wsnL,Rng=True,StLng=False,StLtr=False, 
            rngX=(1e-2,1e3), rngY=(1e-2,1e6))

    """ Test05 : info srimWS, srimWB
    ----.---.---.---.---.---.---.---.---.---.---.---.---"""
    if(SwTest05):
        print('====== Test-05 ======')
        sr.info(mod=0)      # srimWB simple info.
        sr.WBinfo()         # srimWB info(Beam x Trg table)
        sr.WSinfo('srim84Kr_Si',modL=[-1])      # srimWS all info.
        wsnmL = sr.getWSnmL(Beam='Kr',swSort=1) # get WSnames
        sr.WBdump(wsnmL)    # dump srWB selected wsnmL
        #sr.WBdump()         # dump srWB all srWS to Power-Point file        
        
    print('@@main: -- Fin ---')
